#include <stdio.h>
#include <stdlib.h>
#include "gene_model.h"
#include "gencode_gtf.h"
#include "../annotation/annot_region_tools.h"
#include "../../../lib/utils/file/utils.h"
#include "../../../lib/utils/ansi_string/ansi_string.h"
#include "../../../lib/genomics_utils/genome_sequence/genome_sequence_tools.h"
#include "../../../lib/genomics_utils/alignment/alignment_tools.h"
#include <algorithm>

t_gene_list::t_gene_list()
{
	this->gene_list = new vector<t_gene*>();
}

/*
Merge the genes per coordinates in a (non)strand specific way. The merging also stores the pointers to the
genes that are merged for each region.
*/
vector<t_annot_region*>* t_gene_list::get_merged_gene_region_list()
{
	vector<t_annot_region*>* merged_gene_region_list = new vector<t_annot_region*>();

	// Divide the genes wrt the chromsomes they reside on.
	vector<char*>* chr_ids = this->get_chr_ids();

	fprintf(stderr, "Found genes in %d distinct chromosomes.\n", chr_ids->size());

	// Process the genes on each chromosome separately.
	for(int i_chr = 0; i_chr < chr_ids->size(); i_chr++)
	{
		fprintf(stderr, "Merging the genes on %s\n", chr_ids->at(i_chr));

		vector<t_gene*>* cur_chr_genes = this->get_genes_per_chromosome(chr_ids->at(i_chr));

		// Sort the genes on the current chromosome.
		sort(cur_chr_genes->begin(), cur_chr_genes->end(), sort_genes);

		if(cur_chr_genes->size() > 0)
		{
			unsigned int i = 0;
			int cur_start = cur_chr_genes->at(i)->start;
			int cur_end = cur_chr_genes->at(i)->end;

			// Initialize the list of genes that are merged.
			vector<t_gene*>* cur_merged_region_genes = new vector<t_gene*>();
			cur_merged_region_genes->push_back(cur_chr_genes->at(i));

			/*
			Following is copied totally from annot_region_tools.cpp.
			*/
			// Find the next read that is to be merged in this chromosome.
			while(i < cur_chr_genes->size())
			{
				// Is this not the last region in the region list?
				if(i+1 < cur_chr_genes->size())
				{
					if(cur_chr_genes->at(i+1)->start <= cur_end)
					{
						//fprintf(stderr, "Merging: <%d-%d>, cur_end = %d\n", cur_chr_genes->at(i+1)->start, cur_chr_genes->at(i+1)->end, cur_end);
						// Update cur_end, if the end of next region is passing cur_end, update cur_end, otherwise this region is a subset of previous regions.
						cur_end = (cur_end < cur_chr_genes->at(i+1)->end)?(cur_chr_genes->at(i+1)->end):cur_end;

						// Add the currently merged gene to the list of genes.
						cur_merged_region_genes->push_back(cur_chr_genes->at(i+1));
					}
					else
					{
						// Add the region that has accumulated so far: Note that the last region is processed in the previous iteration, it therefore is not 
						// necessary to process that last region any more.
						t_annot_region* new_region = new t_annot_region();
						new_region->start = cur_start;
						new_region->end = cur_end;
						new_region->chrom = new char[strlen(cur_chr_genes->at(i)->chrom) + 2];
						strcpy(new_region->chrom, cur_chr_genes->at(i)->chrom);
						new_region->strand = cur_chr_genes->at(i)->strand;
						new_region->data = (void*)(cur_merged_region_genes);

						merged_gene_region_list->push_back(new_region);
						//fprintf(stderr, "Adding merged region <%d-%d>.\n", merged_gene_region_list->back()->start, merged_gene_region_list->back()->end);

						// Set the current start to start of next region.						
						cur_start = cur_chr_genes->at(i+1)->start;
						cur_end = cur_chr_genes->at(i+1)->end;
						cur_merged_region_genes = new vector<t_gene*>();
						cur_merged_region_genes->push_back(cur_chr_genes->at(i+1));
						//printf("Initiated a new region @ <%d-.>\n", cur_start);
					}
				}
				else
				{
					// This is the last region. Set the max to its ending and add the last merged region.
					t_annot_region* new_region = new t_annot_region();
					new_region->start = cur_start;
					//new_region->end = cur_chr_regions->at(i)->end;
					new_region->end = cur_end;
					new_region->chrom = new char[strlen(cur_chr_genes->at(i)->chrom) + 2];
					strcpy(new_region->chrom, cur_chr_genes->at(i)->chrom);
					new_region->strand = cur_chr_genes->at(i)->strand;
					new_region->data = (void*)(cur_merged_region_genes);

					merged_gene_region_list->push_back(new_region);

					//fprintf(stderr, "Adding merged region <%d-%d>.\n", merged_gene_region_list->back()->start, merged_gene_region_list->back()->end);
					break;
				}

				i++;
			} // i loop 
		} // size check for number of genes on the current chromsome.
	} // i_chr loop.

	for(int i_r = 0; i_r < merged_gene_region_list->size(); i_r++)
	{
		vector<t_gene*>* cur_gene_list = (vector<t_gene*>*)(merged_gene_region_list->at(i_r)->data);
		fprintf(stderr, "%d: %s:%d-%d\n", cur_gene_list->size(), merged_gene_region_list->at(i_r)->chrom, merged_gene_region_list->at(i_r)->start, 
			merged_gene_region_list->at(i_r)->end);

		for(int i_g = 0 ;i_g < cur_gene_list->size(); i_g++)
		{
			fprintf(stderr, "%s: %s:%d-%d\n", cur_gene_list->at(i_g)->gene_id, cur_gene_list->at(i_g)->chrom, cur_gene_list->at(i_g)->start, cur_gene_list->at(i_g)->end);
		} // i_g loop.

		//if(cur_gene_list->size() == 1)
		//{
		//	getc(stdin);
		//}
	} // i_r loop.

	return(merged_gene_region_list);
}

vector<t_gene*>* t_gene_list::get_genes_per_chromosome(char* _chrom)
{
	vector<t_gene*>* chrom_genes = new vector<t_gene*>();
	for(int i_g = 0; i_g < this->gene_list->size(); i_g++)
	{
		if(strcmp(this->gene_list->at(i_g)->chrom, _chrom) == 0)
		{
			chrom_genes->push_back(this->gene_list->at(i_g));
		}
	} // i_g loop

	return(chrom_genes);
}

vector<char*>* t_gene_list::get_chr_ids()
{
	vector<char*>* chr_ids = new vector<char*>();

	for(int i_g = 0; i_g < this->gene_list->size(); i_g++)
	{
		if(chr_ids->size() == 0)
		{
			char* new_chr_id = new char[strlen(this->gene_list->at(i_g)->chrom) + 1];
			strcpy(new_chr_id, this->gene_list->at(i_g)->chrom);
			chr_ids->push_back(new_chr_id);
		}
		else
		{
			int i_chr = t_string::get_i_str(chr_ids, this->gene_list->at(i_g)->chrom);

			// Add the new chromosome id.
			if(i_chr == chr_ids->size())
			{
				char* new_chr_id = new char[strlen(this->gene_list->at(i_g)->chrom) + 1];
				strcpy(new_chr_id, this->gene_list->at(i_g)->chrom);
				chr_ids->push_back(new_chr_id);
			}
		}
	} // i_g loop.

	return(chr_ids);
}

char t_gene_list::get_annotation_signal_per_transcript(int coordinate, t_transcript* transcript)
{
	for(int i_cds = 0; i_cds < transcript->CDSs->size(); i_cds++)
	{
		if(transcript->CDSs->at(i_cds)->start <= coordinate && 
			transcript->CDSs->at(i_cds)->end >= coordinate)
		{
			return(ANNOT_SIGNAL_CDS);
		}
	} // i_e loop.

	for(int i_utr = 0; i_utr < transcript->UTRs->size(); i_utr++)
	{
		if(transcript->UTRs->at(i_utr)->start <= coordinate && 
			transcript->UTRs->at(i_utr)->end >= coordinate)
		{
			return(ANNOT_SIGNAL_UTR);
		}
	} // i_e loop.

	return(ANNOT_SIGNAL_INTRON);
}

void t_gene_list::restructure_gencode_entries(vector<t_gencode_entry*>* gencode_entries)
{
	// Go over all the entries and form the structure based on the biotypes of the features.
	printf("Adding genes.\n");
	for(int i_entry = 0; i_entry < gencode_entries->size(); i_entry++)
	{
		if(i_entry % 1000 == 0)
		{
			fprintf(stderr, "%d. entry.        \r", i_entry);
		}

		// Check the feature biotype for the current entry.
		if(gencode_entries->at(i_entry)->feature_id == gene)
		{
			this->add_gene(gencode_entries->at(i_entry));
		}
	} // i_entry loop.

	printf("Loaded %d genes.\n", this->gene_list->size());

	// Sort gene entries before adding exons/UTRs/start-stop codons.
	sort(this->gene_list->begin(), this->gene_list->end(), sort_genes);

	printf("Adding transcripts.\n");
	for(int i_entry = 0; i_entry < gencode_entries->size(); i_entry++)
	{
		if(i_entry % 1000 == 0)
		{
			fprintf(stderr, "%d. entry.        \r", i_entry);
		}

		// Check the feature biotype for the current entry.
		if(gencode_entries->at(i_entry)->feature_id == transcript)
		{
			this->add_transcript(gencode_entries->at(i_entry));
		}
	} // i_entry loop.

	printf("Adding exons, UTRs, CDSs, start, stop codons.\n");
	for(int i_entry = 0; i_entry < gencode_entries->size(); i_entry++)
	{
		if(i_entry % 1000 == 0)
		{
			fprintf(stderr, "%d. entry.        \r", i_entry);
		}

		// Check the feature biotype for the current entry.
		if(gencode_entries->at(i_entry)->feature_id == exon)
		{
			this->add_exon(gencode_entries->at(i_entry));
		}
		else if(gencode_entries->at(i_entry)->feature_id == UTR)
		{
			this->add_UTR(gencode_entries->at(i_entry));
		}
		else if(gencode_entries->at(i_entry)->feature_id == CDS)
		{
			this->add_CDS(gencode_entries->at(i_entry));
		}
		else if(gencode_entries->at(i_entry)->feature_id == start_codon)
		{
			this->add_start_codon(gencode_entries->at(i_entry));
		}
		else if(gencode_entries->at(i_entry)->feature_id == stop_codon)
		{
			this->add_stop_codon(gencode_entries->at(i_entry));
		}
	} // i_entry loop.

	for(int i_g = 0; i_g < this->gene_list->size(); i_g++)
	{
		t_gene* cur_gene = this->gene_list->at(i_g);

		if(cur_gene->transcripts->size() == 0)
		{
			printf("%s has no transcripts.\n", cur_gene->gene_id);
			//getc(stdin);
		}		

		if(cur_gene->transcripts->size() > 0)
		{
			for(int i_t = 0; i_t < cur_gene->transcripts->size(); i_t++)
			{
				t_transcript* cur_transcript = cur_gene->transcripts->at(i_t);
				sort(cur_transcript->exons->begin(), cur_transcript->exons->end(), sort_exons);
			} // i_t loop.
		} // # transcript check.
	} // i_g loop

	// Restructure the genes to access them easier.
	fprintf(stderr, "Restructuring the gene list.\n");
	this->chrom_ids = get_chr_ids();
	this->genes_per_chromosome = new vector<t_gene*>*[this->chrom_ids->size()];

	for(int i_chrom = 0; i_chrom < this->chrom_ids->size(); i_chrom++)
	{
		this->genes_per_chromosome[i_chrom] = this->get_genes_per_chromosome(this->chrom_ids->at(i_chrom));
	} // i_chrom 

	/*
	// Dump some of the protein coding genes.
	for(int i_g = 0; i_g < this->gene_list->size(); i_g++)
	{
		t_gene* cur_gene = this->gene_list->at(i_g);

		if(cur_gene->transcripts->size() == 0)
		{
			printf("%s has no transcripts.\n", cur_gene->gene_id);
			//getc(stdin);
		}		

		if(cur_gene->gene_type == protein_coding &&
			cur_gene->transcripts->size() > 0)
		{
			for(int i_t = 0; i_t < cur_gene->transcripts->size(); i_t++)
			{
				t_transcript* cur_transcript = cur_gene->transcripts->at(i_t);

				if(cur_transcript->exons->size() > 0 &&
					cur_transcript->transcript_type == protein_coding)
				{
					printf("%s: ", cur_transcript->transcript_id);

					// Sort the exons.
					sort(cur_transcript->exons->begin(), cur_transcript->exons->end(), sort_exons);

					for(int i_e = 0; i_e < cur_transcript->exons->size(); i_e++)
					{
						fprintf(stderr, "(%d-%d) ", cur_transcript->exons->at(i_e)->start, cur_transcript->exons->at(i_e)->end);
					} // i_e loop.

					fprintf(stderr, "\n");
				}
			} // i_t loop.
		}
	} // i_g loop.
	*/
	// Relate the CDSs and UTRs to the exons: For each transcript, assign the CDS's and UTRs to the exons.
}

void t_gene_list::dump_UTRs_per_transcript(vector<char>* gene_types, 
											vector<char>* transcript_types)
{
	// Dump 5' and 3' UTRs for all the transcripts:
	// Go over all the genes and dump the 3' UTRs for each gene, whenever possible. It may be necessary to separate those with respect to the protein coding genes 
	// and others.
	vector<t_annot_region*>* fp_utrs = new vector<t_annot_region*>();
	vector<t_annot_region*>* tp_utrs = new vector<t_annot_region*>();

	for(int i_g = 0; i_g < this->gene_list->size(); i_g++)
	{
		t_gene* current_gene = this->gene_list->at(i_g);

		bool process_current_gene = false;

		if(gene_types != NULL)
		{
			// Check if this gene type is requested.
			for(int i_t = 0; 
				!process_current_gene && i_t < gene_types->size(); 
				i_t++)
			{
				if(current_gene->gene_type == gene_types->at(i_t))
				{
					process_current_gene = true;
				}
			} // i_t loop for gene_type checking.
		} // gene_type check.
		else
		{
			// If the gene types are not spicified, process all the genes.
			process_current_gene = true;
		}

		// Go over all the transcripts of the current gene.
		for(int i_t = 0; 
			process_current_gene && i_t < current_gene->transcripts->size(); 
			i_t++)
		{
			t_transcript* current_transcript = current_gene->transcripts->at(i_t);

			bool process_current_transcript = true;

			if(transcript_types != NULL)
			{
				// Check if this gene type is requested.
			}

			// Is this transcript processable?
			if(process_current_transcript)
			{				
				// Many entries do not have UTRs, for example the ncRNAs.
				if(current_transcript->UTRs->size() > 0)
				{
					// Dump the 5' UTR and 3' UTR: Must have UTRs and must have 2 of them. Do sanity check.
					if(current_transcript->UTRs->size() != 2)
					{
						printf("Transcript %s has %d UTRs associated with it.\n", current_transcript->transcript_id, current_transcript->UTRs->size());
						//exit(0);
					}

					// Sort the UTRs.
					sort(current_transcript->UTRs->begin(), current_transcript->UTRs->end(), sort_regions);

					// If the transcript is on the negative strand, 5' UTR is on the righthand side, 3' UTR is on the left hand side and vice versa.
					if(current_transcript->strand == '-')
					{
						fp_utrs->push_back(current_transcript->UTRs->back());
						tp_utrs->push_back(current_transcript->UTRs->at(0));
					}
					else if(current_transcript->strand == '+')
					{
						fp_utrs->push_back(current_transcript->UTRs->at(0));
						tp_utrs->push_back(current_transcript->UTRs->back());
					}
					else
					{
						printf("Could not recognize the strand character '%c' for transcript %s\n", current_transcript->strand,
							current_transcript->transcript_id);
						exit(0);
					} // Strand checks.
				} // UTR # check.
				else
				{
					printf("Transcript %s has no UTRs associated with it.\n", current_transcript->transcript_id);
				} // UTR # check.
			} // transcript processability check.
		} // i_t loop.
	} // i_g loop.

	printf("Found %d and %d 5' and 3' UTRs.\n", fp_utrs->size(), tp_utrs->size());

	// Dump the utrs.
	dump_BED("fp_utrs.txt", fp_utrs);
	dump_BED("tp_utrs.txt", tp_utrs);
}

vector<t_gene*>* t_gene_list::get_genes_by_gene_type(vector<char>* types)
{
	vector<t_gene*>* gene_list_by_gene_type = new vector<t_gene*>();

	for(int i_g = 0; i_g < this->gene_list->size(); i_g++)
	{
		bool found_entry = false;
		for(int i_t = 0; 
			!found_entry && i_t < types->size(); 
			i_t++)
		{
			if(this->gene_list->at(i_g)->gene_type == types->at(i_t))
			{
				gene_list_by_gene_type->push_back(this->gene_list->at(i_g));
				found_entry = true;
			}
		} // i_t loop.
	} // i_g loop.

	return(gene_list);
}

void t_gene_list::dump_BED_per_gene_list(vector<t_gene*>* _gene_list, char* op_fp)
{
	FILE* f_op = open_f(op_fp, "w");

	for(int i_g = 0; i_g < _gene_list->size(); i_g++)
	{
		fprintf(f_op, "%s\t%d\t%d\t.\t.%c\n", 
			_gene_list->at(i_g)->chrom, 
			_gene_list->at(i_g)->start,
			_gene_list->at(i_g)->end, 
			_gene_list->at(i_g)->strand);
	} // i_g loop.

	fclose(f_op);
}

char* t_gene_list::get_protein_sequence_per_transcript(t_transcript* transcript, char* bin_genome_seq_dir)
{
	// Get the coding transcript from cds's.
	char* total_coding_sequence = new char[transcript->end - transcript->start + 1];
	memset(total_coding_sequence, 0, transcript->end - transcript->start + 1);

	// Sort regions.
	sort(transcript->CDSs->begin(), transcript->CDSs->end(), sort_regions);

	// Make sure that  at the start and stop codons.

	// Process the start codon and the CDS that contains the start codon.
	for(int i = 0; i < transcript->CDSs->size(); i++)
	{
		t_annot_region* current_cds = transcript->CDSs->at(i);

		// Extract the genomic sequence for the current cds, extract the positive strand.
		char* current_cds_seq = extract_sequence(bin_genome_seq_dir, current_cds->chrom, current_cds->start, current_cds->end, '+');

		// Concatenate the cds's.
		strcat(total_coding_sequence, current_cds_seq);		
	} // i loop.

	// Reverse complement the coding sequence if on the negative strand.
	if(transcript->strand == '-')
	{
		get_reverse_complement(total_coding_sequence, 0, strlen(total_coding_sequence) - 1);
	}

	// 

	// Sanity check: Make sure that total coding sequence has multiple of 3 nucleotides in total.
	if(strlen(total_coding_sequence) % 3 != 0)
	{
		fprintf(stderr, "The total length of coding sequence in transcript %s is %d.\n", total_coding_sequence, strlen(total_coding_sequence));
		exit(0);
	}
}

vector<t_transcript*>* t_gene_list::get_transcripts()
{
	vector<t_transcript*>* transcripts = new vector<t_transcript*>();

	for(int i_g = 0; i_g < this->gene_list->size(); i_g++)
	{
		t_gene* cur_gene = this->gene_list->at(i_g);
		for(int i_t = 0; i_t < cur_gene->transcripts->size(); i_t++)
		{
			transcripts->push_back(cur_gene->transcripts->at(i_t));
		} // i_t loop.
	} // i_g loop.

	return(transcripts);
}

vector<t_gene*>* t_gene_list::get_ncRNA_genes()
{
	vector<char>* ncrna_biotypes = new vector<char>();

	/*
Mt_rRNA,
Mt_tRNA,
miRNA,
misc_RNA,
rRNA,
snRNA,
SnoRNA,
	*/
	ncrna_biotypes->push_back(Mt_rRNA);
	ncrna_biotypes->push_back(Mt_tRNA);
	ncrna_biotypes->push_back(tRNAscan);
	ncrna_biotypes->push_back(miRNA);
	ncrna_biotypes->push_back(misc_RNA);
	ncrna_biotypes->push_back(rRNA);
	ncrna_biotypes->push_back(snRNA);
	ncrna_biotypes->push_back(SnoRNA);
	ncrna_biotypes->push_back(lincRNA);
	ncrna_biotypes->push_back(processed_transcript);
	
	vector<t_gene*>* ncrna_genes = this->get_genes_by_gene_type(ncrna_biotypes);

	return(ncrna_genes);
}

bool sort_exons(t_exon* exon1, t_exon* exon2)
{
	return(exon1->start < exon2->start);
}

bool sort_transcripts(t_transcript* tr1, t_transcript* tr2)
{
	return(tr1->start < tr2->start);
}

bool sort_genes(t_gene* gene1, t_gene* gene2)
{
	return(gene1->start < gene2->start);
}

/*
Following does a fast binary search over the starting positions of the regions to find two
genes such that posn is between their start's.

cur_side indicates the relative position of the last middle gene's start with respect to posn.
*/
//int t_gene_list::get_closest_gene(int posn, int& cur_side, int i_start, int i_end)
int t_gene_list::get_closest_gene(int posn, int i_start, int i_end)
{
	if(posn >= this->gene_list->back()->start)
	{
		//cur_side = SIDE_LEFT;
		return(this->gene_list->size() - 1);
	}
	else if(posn <= this->gene_list->at(0)->start)
	{
		return(0);
	}

	if((i_end - i_start) <= 1)
	{
		return(i_end);
	}

	int i_mid = (int)((i_start + i_end) / 2);

	t_gene* gene_at_mid = this->gene_list->at(i_mid);

	if(gene_at_mid->start == posn)
	{
		//cur_side = SIDE_LEFT;
        while(this->gene_list->at(i_mid)->start == posn)
        {
                i_mid++;
        }

        return(i_mid);
	}
	else if(gene_at_mid->start > posn)
	{
		return(this->get_closest_gene(posn, i_start, i_mid));
	}
	else 
	{
		return(this->get_closest_gene(posn, i_mid, i_end));
	}
}

/*
Make sure that the gene list is sorted.
*/
t_gene* t_gene_list::get_gene(char* gene_id, int posn)
{
	int i_gene = this->get_closest_gene(posn, 0, this->gene_list->size() - 1);

	// Go to left to find the gene.
	int i_g = i_gene;
	while(i_g >= 0)
	{
		if(strcmp(this->gene_list->at(i_g)->gene_id, gene_id) == 0)
		{
			return(this->gene_list->at(i_g));
		}

		i_g--;
	}

	return(NULL);
}

t_gene* t_gene_list::get_gene(char* gene_id)
{
	// Check if the gene exists already.
	for(int i_g = 0; i_g < this->gene_list->size(); i_g++)
	{
		if(strcmp(this->gene_list->at(i_g)->gene_id, gene_id) == 0)
		{
			// The gene exsits.
			return(this->gene_list->at(i_g));
		}
	} // i_g loop.
	
	return(NULL);
}

t_transcript* t_gene_list::get_transcript(t_gene* gene, char* _transcript_id)
{
	// Check if the gene exists already.
	for(int i_t = 0; i_t < gene->transcripts->size(); i_t++)
	{
		if(strcmp(gene->transcripts->at(i_t)->transcript_id, _transcript_id) == 0)
		{
			// The gene exsits.
			return(gene->transcripts->at(i_t));
		}
	} // i_g loop.
	
	return(NULL);
}

t_exon* get_exon(t_transcript* transcript, int start, int end)
{
	// Check if the gene exists already.
	for(int i_e = 0; i_e < transcript->exons->size(); i_e++)
	{
		if(transcript->exons->at(i_e)->start == start &&
			transcript->exons->at(i_e)->end == end)
		{
			// The gene exsits.
			return(transcript->exons->at(i_e));
		}
	} // i_g loop.
	
	return(NULL);
}

t_gene* t_gene_list::add_gene(char* _gene_id, char gene_type, char* _chrom, char _strand, int _start, int _end)
{
	if(this->get_gene(_gene_id) != NULL)
	{
		return(NULL);
	}

	// Allocate a new gene entry.
	t_gene* new_gene = new t_gene(_gene_id, gene_type, _chrom, _strand, _start, _end);
	this->gene_list->push_back(new_gene);
	return(new_gene);
}

t_transcript* t_gene_list::add_transcript(char* _gene_id, char* _transcript_id, char transcript_type, char* _chrom, char _strand, int _start, int _end)
{
	t_gene* owner_gene = this->get_gene(_gene_id);
	if(owner_gene == NULL)
	{
		// Add the gene with the coordinates from the transcripts.
		printf("Could not find the gene with %s to add %s\n", _gene_id, _transcript_id);
		exit(0);
		//owner_gene = new t_gene(_gene_id, _chrom, _strand, _start, _end);
		//this->gene_list->push_back(owner_gene);
	}

	// Add the transcript to the transcript list of the owner gene.
	if(this->get_transcript(owner_gene, _transcript_id) != NULL)
	{
		printf("Transcript %s is already assigned to a gene.\n", _transcript_id);
		return(NULL);
	}

	t_transcript* new_transcript = new t_transcript(_gene_id, _transcript_id, transcript_type, _chrom, _strand, _start, _end);
	owner_gene->transcripts->push_back(new_transcript);

	return(new_transcript);
}

t_exon* t_gene_list::add_exon(char* _gene_id, char* _transcript_id, char* _chrom, char _strand, int _start, int _end)
{
	// Get the owner transcript, then add the exon.
	t_gene* owner_gene = this->get_gene(_gene_id);
	if(owner_gene == NULL)
	{
		// Add the gene with the coordinates from the transcripts.
		printf("The owner gene does not exist for transcript %s for adding exon.", _gene_id);
		exit(0);
	}

	// Add the transcript to the transcript list of the owner gene.
	t_transcript* owner_transcript = this->get_transcript(owner_gene, _transcript_id);
	if(owner_transcript == NULL)
	{
		printf("The owner transcript does not exist for transcript %s for adding exon.", _transcript_id);
		exit(0);
	}

	t_exon* new_exon = new t_exon(_gene_id, _transcript_id, _chrom, _strand, _start, _end);

	owner_transcript->exons->push_back(new_exon);

	return(new_exon);
}

t_annot_region* transcript_2_region(t_transcript* transcript)
{
	t_annot_region* new_region = new t_annot_region();
	new_region->chrom = new char[strlen(transcript->chrom) + 2];
	strcpy(new_region->chrom, transcript->chrom);
	new_region->start = transcript->start;
	new_region->end = transcript->end;
	new_region->strand = transcript->strand;

	return(new_region);
}

t_annot_region* gene_2_region(t_gene* gene)
{
	t_annot_region* new_region = new t_annot_region();
	new_region->chrom = new char[strlen(gene->chrom) + 2];
	strcpy(new_region->chrom, gene->chrom);
	new_region->start = gene->start;
	new_region->end = gene->end;
	new_region->strand = gene->strand;

	return(new_region);
}

t_annot_region* exon_2_region(t_exon* exon)
{
	t_annot_region* new_region = new t_annot_region();
	new_region->chrom = new char[strlen(exon->chrom) + 2];
	strcpy(new_region->chrom, exon->chrom);
	new_region->start = exon->start;
	new_region->end = exon->end;
	new_region->strand = exon->strand;

	return(new_region);
}

vector<t_annot_region*>* merge_transcripts(vector<t_transcript*>* transcripts)
{
	vector<t_annot_region*>* exon_regions = new vector<t_annot_region*>();

	// Put all the exons together then merge them.
	for(int i_t = 0; i_t < transcripts->size(); i_t++)
	{	
		t_transcript* cur_transcript = transcripts->at(i_t);

		for(int i_e = 0; i_e < cur_transcript->exons->size(); i_e++)
		{
			exon_regions->push_back(exon_2_region(cur_transcript->exons->at(i_e)));
		} // i_e loop.
	} // i_t loop.

	// Merge the transcript regions.
	vector<t_annot_region*>* merged_transcripts = merge_annot_regions(exon_regions, 0, true);

	// Free memory.
	delete_annot_regions(exon_regions);

	return(merged_transcripts);
}

t_gene::t_gene(char* _gene_id, char _gene_type, char* _chrom, char _strand, int _start, int _end)
{
	this->gene_id = new char[strlen(_gene_id) + 2];
	strcpy(this->gene_id, _gene_id);
	
	this->chrom = new char[strlen(_chrom) + 2];
	strcpy(this->chrom, _chrom);

	this->strand = _strand;
	this->start = _start;
	this->end = _end;

	this->gene_type = _gene_type;

	this->transcripts = new vector<t_transcript*>();

	// Set the user specific data to NULL.
	this->data = NULL;
}

void t_transcript::get_promoter(int l_ext, char* prom_chrom, int& prom_start, int& prom_end)
{
	if(this->strand == '+')
	{
		prom_start = this->start - l_ext;
		prom_end = this->start + l_ext;
		strcpy(prom_chrom, this->chrom);
	}
	else
	{
		prom_start = this->end - l_ext;
		prom_end = this->end + l_ext;
		strcpy(prom_chrom, this->chrom);
	}
}

t_transcript::t_transcript(char* _gene_id, 
							char* _transcript_id, 
							char _transcript_type, 
							char* _chrom, 
							char _strand, 
							int _start, 
							int _end)
{
	this->gene_id = new char[strlen(_gene_id) + 2];
	strcpy(this->gene_id, _gene_id);

	this->transcript_id = new char[strlen(_transcript_id) + 2];
	strcpy(this->transcript_id, _transcript_id);

	this->chrom = new char[strlen(_chrom) + 2];
	strcpy(this->chrom, _chrom);

	this->strand = _strand;
	this->start = _start;
	this->end = _end;

	this->start_codon = NULL;
	this->stop_codon = NULL;

	this->CDSs = new vector<t_annot_region*>();
	this->UTRs = new vector<t_annot_region*>();

	this->exons = new vector<t_exon*>();

	this->transcript_type = _transcript_type;

	// Set the data to nothing.
	this->data = NULL;
}

t_exon::t_exon(char* _gene_id, char* _transcript_id, char* _chrom, char _strand, int _start, int _end)
{
	this->chrom = new char[strlen(_chrom) + 2];
	strcpy(this->chrom, _chrom);
	this->gene_id = new char[strlen(_gene_id) + 2];
	strcpy(this->gene_id, _gene_id);
	this->transcript_id = new char[strlen(_transcript_id) + 2];
	strcpy(this->transcript_id, _transcript_id);

	this->start = _start;
	this->end = _end;
	this->strand = _strand;
}

t_gene* t_gene_list::add_gene(t_gencode_entry* _entry)
{
	t_gene* new_gene = this->get_gene(_entry->gene_id);
	if(new_gene != NULL)
	{
		printf("The new gene %s already exists.", _entry->gene_id);
		exit(0);
	}

	new_gene = new t_gene(_entry->gene_id, 
							_entry->gene_type,
							chromosomes[_entry->chr_id], 							
							_entry->strand, 
							_entry->start, 
							_entry->end);

	this->gene_list->push_back(new_gene);

	return(new_gene);
}

t_transcript* t_gene_list::add_transcript(t_gencode_entry* _entry)
{
	t_gene* owner_gene = this->get_gene(_entry->gene_id, _entry->start);
	if(owner_gene == NULL)
	{
		printf("The owner gene %s does not exist for transcript %s.", _entry->gene_id, _entry->transcript_id);
		exit(0);
	}

	t_transcript* new_transcript = new t_transcript(_entry->gene_id, _entry->transcript_id, _entry->transcript_type, chromosomes[_entry->chr_id], _entry->strand, _entry->start, _entry->end);

	owner_gene->transcripts->push_back(new_transcript);

	return(new_transcript);
}

t_exon* t_gene_list::add_exon(t_gencode_entry* _entry)
{
	// Get the owner transcript, then add the exon.
	t_gene* owner_gene = this->get_gene(_entry->gene_id, _entry->start);
	//t_gene* owner_gene = this->get_gene(_entry->gene_id);
	if(owner_gene == NULL)
	{
		// Add the gene with the coordinates from the transcripts.
		printf("The owner gene does not exist for transcript %s for adding exon.", _entry->gene_id);
		exit(0);
	}

	// Add the transcript to the transcript list of the owner gene.
	t_transcript* owner_transcript = this->get_transcript(owner_gene, _entry->transcript_id);
	if(owner_transcript == NULL)
	{
		printf("The owner transcript does not exist for transcript %s for adding exon.", _entry->transcript_id);
		exit(0);
	}

	t_exon* new_exon = new t_exon(_entry->gene_id, _entry->transcript_id, chromosomes[_entry->chr_id], _entry->strand, _entry->start, _entry->end);

	owner_transcript->exons->push_back(new_exon);

	return(new_exon);
}

t_annot_region* t_gene_list::add_CDS(t_gencode_entry* _entry)
{
	// Get the owner transcript, then add the exon.
	t_gene* owner_gene = this->get_gene(_entry->gene_id, _entry->start);
	if(owner_gene == NULL)
	{
		// Add the gene with the coordinates from the transcripts.
		printf("The owner gene does not exist for transcript %s for adding exon.", _entry->gene_id);
		exit(0);
	}

	// Add the transcript to the transcript list of the owner gene.
	t_transcript* owner_transcript = this->get_transcript(owner_gene, _entry->transcript_id);
	if(owner_transcript == NULL)
	{
		printf("The owner transcript does not exist for transcript %s for adding exon.", _entry->transcript_id);
		exit(0);
	}

	//t_exon* new_exon = new t_exon(_entry->gene_id, _entry->transcript_id, chromosomes[_entry->chr_id], _entry->strand, _entry->start, _entry->end);
	t_annot_region* new_cds = new t_annot_region();
	new_cds->chrom = new char[strlen(chromosomes[_entry->chr_id]) + 2];
	strcpy(new_cds->chrom, chromosomes[_entry->chr_id]);
	new_cds->start = _entry->start;
	new_cds->end = _entry->end;
	new_cds->strand = _entry->strand;

	owner_transcript->CDSs->push_back(new_cds);

	return(new_cds);
}

t_annot_region* t_gene_list::add_UTR(t_gencode_entry* _entry)
{
	// Get the owner transcript, then add the exon.
	t_gene* owner_gene = this->get_gene(_entry->gene_id, _entry->start);
	if(owner_gene == NULL)
	{
		// Add the gene with the coordinates from the transcripts.
		printf("The owner gene does not exist for transcript %s for adding exon.", _entry->gene_id);
		exit(0);
	}

	// Add the transcript to the transcript list of the owner gene.
	t_transcript* owner_transcript = this->get_transcript(owner_gene, _entry->transcript_id);
	if(owner_transcript == NULL)
	{
		printf("The owner transcript does not exist for transcript %s for adding exon.", _entry->transcript_id);
		exit(0);
	}

	//t_exon* new_exon = new t_exon(_entry->gene_id, _entry->transcript_id, chromosomes[_entry->chr_id], _entry->strand, _entry->start, _entry->end);
	t_annot_region* new_utr = new t_annot_region();
	new_utr->chrom = new char[strlen(chromosomes[_entry->chr_id]) + 2];
	strcpy(new_utr->chrom, chromosomes[_entry->chr_id]);
	new_utr->start = _entry->start;
	new_utr->end = _entry->end;
	new_utr->strand = _entry->strand;

	owner_transcript->UTRs->push_back(new_utr);

	return(new_utr);
}

t_annot_region* t_gene_list::add_start_codon(t_gencode_entry* _entry)
{
	// Get the owner transcript, then add the exon.
	t_gene* owner_gene = this->get_gene(_entry->gene_id, _entry->start);
	if(owner_gene == NULL)
	{
		// Add the gene with the coordinates from the transcripts.
		printf("The owner gene does not exist for transcript %s for adding exon.", _entry->gene_id);
		exit(0);
	}

	// Add the transcript to the transcript list of the owner gene.
	t_transcript* owner_transcript = this->get_transcript(owner_gene, _entry->transcript_id);
	if(owner_transcript == NULL)
	{
		printf("The owner transcript does not exist for transcript %s for adding exon.", _entry->transcript_id);
		exit(0);
	}

	//t_exon* new_exon = new t_exon(_entry->gene_id, _entry->transcript_id, chromosomes[_entry->chr_id], _entry->strand, _entry->start, _entry->end);
	t_annot_region* new_start_codon = new t_annot_region();
	new_start_codon->chrom = new char[strlen(chromosomes[_entry->chr_id]) + 2];
	strcpy(new_start_codon->chrom, chromosomes[_entry->chr_id]);
	new_start_codon->start = _entry->start;
	new_start_codon->end = _entry->end;
	new_start_codon->strand = _entry->strand;

	owner_transcript->start_codon = new_start_codon;

	return(new_start_codon);
}

t_annot_region* t_gene_list::add_stop_codon(t_gencode_entry* _entry)
{
	// Get the owner transcript, then add the exon.
	t_gene* owner_gene = this->get_gene(_entry->gene_id, _entry->start);
	if(owner_gene == NULL)
	{
		// Add the gene with the coordinates from the transcripts.
		printf("The owner gene does not exist for transcript %s for adding exon.", _entry->gene_id);
		exit(0);
	}

	// Add the transcript to the transcript list of the owner gene.
	t_transcript* owner_transcript = this->get_transcript(owner_gene, _entry->transcript_id);
	if(owner_transcript == NULL)
	{
		printf("The owner transcript does not exist for transcript %s for adding exon.", _entry->transcript_id);
		exit(0);
	}

	//t_exon* new_exon = new t_exon(_entry->gene_id, _entry->transcript_id, chromosomes[_entry->chr_id], _entry->strand, _entry->start, _entry->end);
	t_annot_region* new_stop_codon = new t_annot_region();
	new_stop_codon->chrom = new char[strlen(chromosomes[_entry->chr_id]) + 2];
	strcpy(new_stop_codon->chrom, chromosomes[_entry->chr_id]);
	new_stop_codon->start = _entry->start;
	new_stop_codon->end = _entry->end;
	new_stop_codon->strand = _entry->strand;

	owner_transcript->stop_codon = new_stop_codon;

	return(new_stop_codon);
}
