#ifndef __GENE_MODEL__
#define __GENE_MODEL__

#include <vector>

using namespace std;

/*
Encapsulation of human gene model:
Genes -> Transcripts -> {exons->{CDS|UTR|start_codon|stop_codon}}
*/

class t_gene;
class t_transcript;
struct t_annot_region;
class t_exon;
struct t_gencode_entry;

//enum{ANNOT_SIGNAL_UTR, ANNOT_SIGNAL_CDS, ANNOT_SIGNAL_INTRON, ANNOT_SIGNAL_EXON};
enum{ANNOT_SIGNAL_INTRON, ANNOT_SIGNAL_EXON, ANNOT_SIGNAL_UTR, ANNOT_SIGNAL_CDS};


/*
There are two sets of functions; one is for adding elements with the explicit values, second set is for adding elements 
by gencode entries for restructuring the gencode entries.
*/
class t_gene_list
{
public:
	t_gene_list();
	~t_gene_list();

	vector<t_gene*>* gene_list; // The list of genes.
	vector<char*>* chrom_ids;
	vector<t_gene*>** genes_per_chromosome;

	// Add elements by the explicit coordinates/chromosomes.
	t_gene* add_gene(char* _gene_id, char _gene_type, char* _chrom, char _strand, int _start, int _end);
	t_transcript* add_transcript(char* _gene_id, char* _transcript_id, char _transcript_type, char* _chrom, char _strand, int _start, int _end);
	t_exon* add_exon(char* _gene_id, char* _transcript_id, char* _chrom, char _strand, int _start, int _end);
	t_annot_region* add_CDS(char* _gene_id, char* _transcript_id, char* _chrom, char _strand, int _start, int _end);
	t_annot_region* add_UTR(char* _gene_id, char* _transcript_id, char* _chrom, char _strand, int _start, int _end);

	void dump_UTRs_per_transcript(vector<char>* gene_types, 
									vector<char>* transcript_types);

	// Add elements by gencode entries.
	void restructure_gencode_entries(vector<t_gencode_entry*>* gencode_entries);
	t_gene* add_gene(t_gencode_entry* _entry);
	t_transcript* add_transcript(t_gencode_entry* _entry);
	t_exon* add_exon(t_gencode_entry* _entry);
	t_annot_region* add_CDS(t_gencode_entry* _entry);
	t_annot_region* add_UTR(t_gencode_entry* _entry);
	t_annot_region* add_start_codon(t_gencode_entry* _entry);
	t_annot_region* add_stop_codon(t_gencode_entry* _entry);

	t_gene* get_gene(char* id);
	t_transcript* get_transcript(t_gene* gene, char* transcript_id);
	t_exon* get_exon(t_transcript* transcript, int start, int end);

	vector<char*>* get_chr_ids();
	vector<t_annot_region*>* get_merged_gene_region_list();
	vector<t_gene*>* get_genes_per_chromosome(char* _chrom);

	static char get_annotation_signal_per_transcript(int coordinate, t_transcript* transcript);

	//int get_closest_gene(int posn, int& cur_side, int i_start, int i_end);
	int get_closest_gene(int posn, int i_start, int i_end);
	t_gene* get_gene(char* gene_id, int posn);

	// Get several regions of interest.
	vector<t_gene*>* get_ncRNA_genes(); // gene_type 
	vector<t_gene*>* get_protein_coding_genes(); // gene_type and transcript_type protein_coding.
	vector<t_gene*>* get_pseudogenes(); // gene_type pseudogene?
	vector<t_annot_region*>* get_introns(); // Do transcript level processing to get the regions between exons. (Strand issues?)
	vector<t_annot_region*>* get_intergenic_regions(); // Do gene level processing to get the regions between genes. (Strand issues?)
	vector<t_annot_region*>* get_transcript_start_sites(); // Do transcript level processing to get the starting positions of the transcripts
	vector<t_annot_region*>* get_non_overlapping_genes(); // Get the list of genes that do not overlap with any other gene on either strand.

	vector<t_transcript*>* get_transcripts();

	char* get_protein_sequence_per_transcript(t_transcript* transcript, char* bin_genome_seq_dir);

	vector<t_gene*>* get_genes_by_gene_type(vector<char>* types);

	static void dump_BED_per_gene_list(vector<t_gene*>* _gene_list, char* op_fp);
};

vector<char*>* get_chromosomes_per_transcripts(vector<t_transcript*>* transripts);
vector<t_transcript*>* get_transcripts_per_chromosome(vector<t_transcript*>* transripts, char* chr);

vector<t_annot_region*>* merge_transcripts(vector<t_transcript*>* transcripts);

// Convert gene units to annot_region.
t_annot_region* gene_2_region(t_gene* gene);
t_annot_region* transcript_2_region(t_transcript* transcript);
t_annot_region* exon_2_region(t_exon* exon);

class t_gene
{
public:
	t_gene(char* _gene_id, char _gene_type, char* _chrom, char _strand, int _start, int _end);

	char* chrom;
	char strand;
	int start;
	int end;
	char* gene_id; 
	char gene_type;
	vector<t_transcript*>* transcripts;

	void* data; // Application specific data: Can be expression values, transcript information, etc.
};

class t_transcript
{
public:
	t_transcript(char* _gene_id, char* _transcript_id, char _transcript_type, char* _chrom, char _strand, int _start, int _end);

	char* chrom;
	char strand;
	int start;
	int end;
	char* gene_id;
	char* transcript_id;
	char transcript_type;
	vector<t_exon*>* exons;

	// Translating parts of an exon: One (not all) can be NULL.
	t_annot_region* start_codon;
	t_annot_region* stop_codon;

	vector<t_annot_region*>* CDSs;
	vector<t_annot_region*>* UTRs;

	// Get the promoter region by extending the gene start l_ext.
	void get_promoter(int l_ext, char* chrom, int& start, int& end);

	void* data; // This is a user definable data which can be set to anything.
};

class t_exon
{
public:
	t_exon(char* _gene_id, char* _transcript_id, char* _chrom, char _strand, int _start, int _end);
	~t_exon();

	char* chrom;
	int start;
	int end;
	char strand;
	char* gene_id;
	char* transcript_id;

	// The coding part of this exon.
	t_annot_region* CDS;

	// UTR part of this exon, if exists.
	t_annot_region* UTR;
};

bool sort_exons(t_exon* exon1, t_exon* exon2);
bool sort_transcripts(t_transcript* tr1, t_transcript* tr2);
bool sort_genes(t_gene* gene1, t_gene* gene2);

#endif // __GENE_MODEL__


